// Glyph Table
// Copyright 1997 Herbert Duerr

#include "xfstt.h"
#include <stdio.h>

GlyphTable::GlyphTable( RandomAccessFile& f, int offset, int length)
:	RandomAccessFile( f, offset, length)
{
	endPoints = 0;
	points = 0;
}

//GlyphTable::~GlyphTable()
//{
//	delete[] endPoints;
//	delete[] points;
//}

void GlyphTable::getGlyphData( int glyphOffset, TTFont* ttFont)
{
	seekAbsolute( glyphOffset);

	nEndPoints = readSShort();

	xmin = readSShort();
	ymin = readSShort();
	xmax = readSShort();
	ymax = readSShort();

	if( nEndPoints < 0) {	// composite glyph
		doCompositeGlyphs( ttFont);
		return;
	}

	endPoints = new int[ nEndPoints];
	for( int iEndPoints = 0; iEndPoints < nEndPoints; ++iEndPoints)
		endPoints[ iEndPoints] = readUShort();

	codeLength = readUShort();
	codeOffset = tellRel();
	seekRelative( codeLength);

	nPoints = 1 + endPoints[ nEndPoints-1];
	points = new point[ nPoints + 2];

	for( int iFlags = 0; iFlags < nPoints; ++iFlags) {
		byte flag = readUByte();
		if( (flag & F_SAME) != 0) {
			for( int j = iFlags + readUByte(); iFlags < j; ++iFlags)
				points[ iFlags].flags = flag;
		}
		points[ iFlags].flags = flag;
	}

	short oldval = 0;
	short delta = 0;
	for( int iXpos = 0; iXpos < nPoints; ++iXpos) {
		switch( points[ iXpos].flags & (X_SHORT | X_EXT)) {
		case 0:	// delta vector
			delta = readSShort();
			break;
		case X_EXT: // same value
			delta = 0;
			break;
		case X_SHORT:
			delta = -readUByte();
			break;
		case X_SHORT | X_EXT:
			delta = +readUByte();
			break;
		}
		oldval += delta;
		points[ iXpos].xnow = oldval;
	}

	oldval = 0;
	for( int iYpos = 0; iYpos < nPoints; ++iYpos) {
		switch( points[ iYpos].flags & (Y_SHORT | Y_EXT)) {
		case 0:	// delta vector
			delta = readSShort();
			break;
		case Y_EXT: // same value
			delta = 0;
			break;
		case Y_SHORT:
			delta = -readUByte();
			break;
		case Y_SHORT | Y_EXT:
			delta = +readUByte();
			break;
		}
		oldval  += delta;
		points[ iYpos].ynow = oldval;
	}
}

void GlyphTable::doCompositeGlyphs( TTFont* ttFont)
{
	GlyphTable *first, *last = 0;
	short flag;

	// read and process subGlyphs
	do {
		GlyphTable* g = new GlyphTable( *this);
		if( last == 0)
			first = g;
		else
			last->next = g;
		last = g;
		g->next = 0;

		flag = readUShort();
		int glyphNo = readUShort();

		ttFont->getGlyphData( glyphNo, g);

		// move and scale

		short arg1, arg2;
		if( (flag & ARGS_ARE_WORDS) != 0) {
			arg1 = readSShort();
			arg2 = readSShort();
		} else {
			arg1 = readSByte();
			arg2 = readSByte();
		}

		int xscale = 0x2000;
		int yscale = 0x2000;
		int scale01 = 0;
		int scale10 = 0;
		if( (flag & HAS_A_SCALE) != 0) {
			xscale = yscale = readSShort();
		} else if( (flag & HAS_XY_SCALE) != 0) {
			xscale = readSShort();
			yscale = readSShort();
		} else if( (flag & HAS_2X2_SCALE) != 0) {
			xscale = readSShort();
			scale01 = readSShort();
			yscale = readSShort();
			scale10 = readSShort();
		}

		// match and/or move
		if( flag & ARGS_ARE_XY) {
			//### move it
			dprintf2( "args_xy = %d %d\n", arg1, arg2);
		} else {
			// match points
			arg1 -= g->points[ 0].xnow;
			arg2 -= g->points[ 0].ynow;
		}

		for( int i = g->nPoints; --i >= 0;) {
			g->points[ i].xnow += arg1;
			g->points[ i].ynow += arg2;
		}

		// scale
		if( xscale != 0x2000)
			for( int i = g->nPoints; --i >= 0;)
				g->points[ i].xnow =
					(xscale * g->points[ i].xnow) >> 14;
		if( yscale != 0x2000)
			for( int i = g->nPoints; --i >= 0;)
				g->points[ i].ynow =
					(yscale * g->points[ i].ynow) >> 14;
		if( scale01 != 0 || scale10 != 0)
			for( int i = g->nPoints; --i >= 0;) {
				int xnow = g->points[ i].xnow;
				int ynow = g->points[ i].ynow;
				g->points[ i].xnow += (scale01 * ynow) >> 14;
				g->points[ i].ynow += (scale01 * xnow) >> 14;
			}
	} while( (flag & MORE_COMPONENTS) != 0);

	if( (flag & HAS_CODE) != 0) {
		codeLength = readUShort();
		codeOffset = tellRel();
		seekRelative( codeLength);
	} else
;//###
		codeLength = 0;

	// prepare for merge
	nEndPoints = 0;
	nPoints = 0;
	for( GlyphTable* g1 = first; g1; g1 = g1->next) {
		nEndPoints += g1->nEndPoints;
		nPoints += g1->nPoints;
	}

	endPoints	= new int[ nEndPoints];
	points		= new point[ nPoints + 2];

	// do the merge
	int epidx = 0;
	int xpidx = 0;
	for( GlyphTable* g2 = first; g2; g2 = last) {
		for( int i21 = 0; i21 < g2->nEndPoints; ++i21, ++epidx)
			endPoints[ epidx] = xpidx + g2->endPoints[ i21];
		for( int i22 = 0; i22 < g2->nPoints; ++i22, ++xpidx)
			points[ xpidx] = g2->points[ i22];
		last = g2->next;
		delete g2;
	}
}

