// Utilities for efficient access to the TTFfile
// Copyright 1997 Herbert Duerr

#include "xfstt.h"
#include <stdio.h>

#ifdef WIN32
	#include <windows.h>
	#include <io.h>
#else
	#include <fcntl.h>
	#include <sys/mman.h>
	#include <sys/stat.h>
	#include <unistd.h>
#endif

#include <stddef.h>
#include <stdlib.h>

void* allocMem( int size)
{
	void* ptr;
#ifdef WIN32
	ptr = VirtualAlloc( NULL, size,
			MEM_COMMIT, PAGE_READWRITE);
#else
	ptr = mmap( 0, size, PROT_READ | PROT_WRITE,
		MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
#endif
	return ptr;
}

void deallocMem( void* ptr, int size)
{
#ifdef WIN32
	VirtualFree( ptr, size, MEM_DECOMMIT);
#else
	munmap( ptr, size);
#endif
}

#ifdef MEMDEBUG

#define MEMPTRS 8192
static void* memptr[ MEMPTRS];
static int memidx = 0;

void cleanupMem()
{
	dprintf0( "Memory holes:\n");
	for( int i = 0; i < memidx; ++i)
		if( memptr[ i])
			dprintf1( "MEM hole = %p\n", memptr[ i]);
}

void* operator new[]( size_t size)
{
	void* ptr = malloc( size);
	dprintf2( "MEM new( %d) = %p\n", size, ptr);

	int i = memidx;
	while( --i >= 0)
		if( memptr[ i] == 0) {
			memptr[ i] = ptr;
			break;
		}
	if( i <= 0)
		memptr[ memidx++] = ptr;

	return ptr;
}

void operator delete[]( void* ptr)
{
	dprintf1( "MEM delete( %p)\n", ptr);

	int i = memidx;
	while( --i >= 0)
		if( memptr[ i] == ptr) {
			memptr[ i] = 0;
			break;
		}
	if( i == memidx - 1)
		--memidx;

	free( ptr);
}

#endif /* MEMDEBUG */


U8* RandomAccessFile::absbase = 0;

RandomAccessFile::RandomAccessFile( char* fileName) {
#ifdef WIN32
	void* fd = CreateFile( fileName, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if( fd == INVALID_HANDLE_VALUE) {
		fprintf( stderr, "Cannot open \"%s\"\n", fileName);
		exit( -1);
	}
	length = GetFileSize( fd, NULL);
	HANDLE hdl = CreateFileMapping( fd, NULL, PAGE_READONLY, 0, length, "TTF mapping");
	base = (byte*) MapViewOfFile( hdl, FILE_MAP_READ, 0, 0, length);
	CloseHandle( fd);
#else
	int fd = open( fileName, O_RDONLY);
	if( fd <= 0) {
		fprintf( stderr, "Cannot open \"%s\"\n", fileName);
		exit( fd);
	}
	struct stat st;
	fstat( fd, &st);
	length = st.st_size;
	base = (byte*) mmap( 0L, length, PROT_READ, MAP_SHARED, fd, 0L);
	close( fd);
#endif
	ptr = base;
	absbase = base;
}

void RandomAccessFile::closeRAFile()
{
#ifdef WIN32
	UnmapViewOfFile( base);
#else
	munmap( base, length);
#endif
}

U32 RandomAccessFile::calcChecksum()
{
	U8* oldptr = ptr;
	int checksum = 0;
	for( int len = length >> 2; --len >= 0;)
		checksum += readUInt();
	if( length & 3)
		checksum += readUInt() & (-1 << ((-length & 3) << 3));
	ptr = oldptr;
dprintf1( "checksum is %08X\n", calcChecksum());
	return checksum;
}

