// Hint Interpreter
// Copyright 1997 Herbert Duerr

#include "xfstt.h"
#include <math.h>	// we need sqrt()

#define FSHIFT 1.0

void Rasterizer::initInterpreter()
{
	endInterpreter();

	if( !grid_fitting || !ttFont)
		return;

	nPointsZ0 = ttFont->maxpTable->maxTwilightPoints;
	p[0] = new point[ nPointsZ0];

	point nullpoint = { 0, 0, 0, 0, 0, 0, 0};
	for( int i = 0; i < nPointsZ0; ++i)
		p[0][i] = nullpoint;

	cvt = new int[ ttFont->cvtTable->numVals()];
	stackbase = new int[ ttFont->maxpTable->maxStackSize];
	stack = stackbase;
	stor = new int[ ttFont->maxpTable->maxStorage];

	fdefs = new FDefs[ ttFont->maxpTable->maxFunctionDefs];
	idefs = new FDefs[ ttFont->maxpTable->maxInstructionDefs];

#ifdef DEBUG
	setbuffer( stdout, 0, 0);
#endif
	if( ttFont->fpgmTable) {
		gs.init( p);
		stack = stackbase;
		execHints( ttFont->fpgmTable, 0, ttFont->fpgmTable->getLength());
	}
}

void Rasterizer::endInterpreter()
{
	if( !stackbase)
		return;

	delete[] idefs;
	delete[] fdefs;
	delete[] stor;
	delete[] stackbase;
	delete[] cvt;
	delete[] p[0];

	stackbase = 0;
}

void Rasterizer::calcCVT()
{
	CvtTable* cvtTab = ttFont->cvtTable;
	if( cvtTab == 0)
		return;

	cvtTab->reset();
	int n = cvtTab->numVals();

	int scale = xx + yy;
	for( int i = 0; i < n; ++i) {
		int val = cvtTab->nextVal();
		dprintf2( "cvt[%3d] = %5d  ", i, val);
		cvt[ i] = ((val * scale + 32) >> 6) << xxexp;
		dprintf1( "-> %5d\n", cvt[ i]);
	}

	if( ttFont->prepTable == 0)
		return;

	dprintf0( "=== starting glyph prep ===\n");
	gs.init( p);
	int preplen = ttFont->prepTable->getLength();
	stack = stackbase;
	execHints( ttFont->prepTable, 0, preplen);
	default_gs = gs;
	dprintf0( "=== glyph prep done ===\n");
}

void GraphicsState::init( point** p)
{
	p_vector_x	= UNITY2D14;
	p_vector_y	= 0;
	f_vector_x	= UNITY2D14;
	f_vector_y	= 0;
	dp_vector_x	= UNITY2D14;
	dp_vector_y	= 0;
	zp2 = zp1 = zp0	= p[1];
	rp2 = rp1 = rp0	= 1;
	loop		= 1;
	auto_flip	= TRUE;
	cvt_cut_in	= (17 << SHIFT) >> 4;
	round_state	= ROUND_GRID;
	round_phase	= 0;
	round_period	= (1 << SHIFT);
	round_thold	= 0;
	min_distance	= (1 << SHIFT);
	swidth_cut_in	= 0;
	swidth_value	= 0;
	delta_base	= 9;
	delta_shift	= 3;
	instr_control	= 0;
	scan_control	= 0;
}

// some inlined utilities

inline int Rasterizer::round( int x)
{
	int y = (x<0) ? -x : x;
	switch( gs.round_state) {
	case ROUND_OFF:
		return x;
	case ROUND_GRID:
		y = (y + 32) & -64;
		break;
	case ROUND_UP:
		y = (y + 63) & -64;
		break;
	case ROUND_DOWN:
		y = y & -64;
		break;
	case ROUND_HALF:
		y = (y & -64) + 32;
		break;
	case ROUND_DOUBLE:
		y = (y + 16) & -32;
		break;
	case ROUND_SUPER:
		y = (y - gs.round_phase + gs.round_thold) & -gs.round_period;
		break;
	case ROUND_SUPER45:
		y = (y - gs.round_phase + gs.round_thold) / gs.round_period;
		y *= gs.round_period;
		break;
	}
	if( y<0) return 0;
	return (x<0) ? -y : y;
}

inline int Rasterizer::absNewMeasure( int dx11D6, int dy11D6)
{
	dprintf2( "\ndx = %d, dy = %d", dx11D6, dy11D6);
	dprintf2( ",\tpx = %d, py = %d", gs.p_vector_x, gs.p_vector_y);

	int dist = dx11D6 * gs.p_vector_x + dy11D6 * gs.p_vector_y + (1 << 13);
	dist >>= 14;
	return dist;
}

inline int Rasterizer::absOldMeasure( int dx11D6, int dy11D6)
{
	dprintf2( "\ndx = %d, dy = %d", dx11D6, dy11D6);
	dprintf2( ",\tdpx = %d, dpy = %d", gs.dp_vector_x, gs.dp_vector_y);

	int dist = dx11D6 * gs.dp_vector_x + dy11D6 * gs.dp_vector_y + (1 << 13);
	dist >>= 14;
	return dist;
}

inline int Rasterizer::newMeasure( point& p2, point& p1)
{
	int dist = absNewMeasure( p2.xnow - p1.xnow, p2.ynow - p1.ynow);
	dprintf3( "\nnewMeasure p[%d]-p[%d] = %f",
		&p2-p[1], &p1-p[1], dist/FSHIFT);
	
	return dist;
}

inline int Rasterizer::oldMeasure( point& p2, point& p1)
{
	int dist = absOldMeasure( p2.xold - p1.xold, p2.yold - p1.yold);
	dprintf3( "\noldMeasure p[%d]-p[%d] = %f",
		&p2-p[1], &p1-p[1], dist/FSHIFT);
	return dist;
}

inline void Rasterizer::newLine2vector( point& p2, point& p1, int& vx, int &vy)
{
	//### how can it be that df1 or df2 are bigger than 16 bits ?
	float f1 = p2.xnow - p1.xnow;
	float f2 = p2.ynow - p1.ynow;
	float f3 = sqrt( UNITY2D14*UNITY2D14 / (f1*f1 + f2*f2));
	vx = (int)(f1 * f3);
	vy = (int)(f2 * f3);

	dprintf2( "(%d %d) - ", p2.xnow, p2.ynow);
	dprintf2( "(%d %d)", p1.xnow, p1.ynow);
	dprintf2( "\nvx vy = %f %f", vx/FSHIFT, vy/FSHIFT);
}

inline void Rasterizer::oldLine2vector( point& p2, point& p1, int& vx, int &vy)
{
	//### how can it be that df1 or df2 are bigger than 16 bits ?
	float f1 = p2.xold - p1.xold;
	float f2 = p2.yold - p1.yold;
	float f3 = sqrt( UNITY2D14*UNITY2D14 / (f1*f1 + f2*f2));
	vx = (int)(f1 * f3);
	vy = (int)(f2 * f3);

	dprintf2( "(%d %d) - ", p2.xold, p2.yold);
	dprintf2( "(%d %d)", p1.xold, p1.yold);
	dprintf2( "\nvx vy = %f %f", vx/FSHIFT, vy/FSHIFT);
}


// interpreter

inline void Rasterizer::execOpcode( int opc, RandomAccessFile* f)
{
	// optimisation question: how do I convince
	// g++ to keep the "this" pointer in a register?
	// all that is needed to make hinting really fast is
	// keeping these variables in registers
	// 1) this 2) m 3) n,
	// keeping i and pp in registers would help too.
	//
	// While I'm at it: Why the heck does g++ take a stack frame
	// of about 1 kB though the function needs much less?
	// Unoptimized MS VC++ code only takes 128 bytes!

	register int m, n, i;
	point* pp;
	int j;

	switch( opc) {

	// pushing onto the stack

	case NPUSHB:
		opc = f->readUByte() + PUSHB00 - 1;
		// fall through
	case PUSHB00: case PUSHB01:
	case PUSHB02: case PUSHB03:
	case PUSHB04: case PUSHB05:
	case PUSHB06: case PUSHB07:
		m = opc - PUSHB00 + 1;
		dprintf1( "PUSHB * %d\n>\t\t", m);
		while( --m >= 0) {
			*(++stack) = f->readUByte();
			dprintf1( "%d,\t", *stack);
			if( (m&7) == 0) dprintf0( "\n>\t\t");
		}
		break;

	case NPUSHW:
		opc = f->readUByte() + PUSHW00 - 1;
		// fall through
	case PUSHW00: case PUSHW01:
	case PUSHW02: case PUSHW03:
	case PUSHW04: case PUSHW05:
	case PUSHW06: case PUSHW07:
		m = opc - PUSHW00 + 1;
		dprintf1( "PUSHW * %d\n>\t\t", m);
		while( --m >= 0) {
			*(++stack) = f->readSShort();
			dprintf1( "%d,\t", *stack);
			if( (m&7) == 0) dprintf0( "\n>\t\t");
		}
		break;

	// accessing the storage area

	case RS:
		m = *stack;
		n = stor[ m];
		*stack = n;
		dprintf2( "RS store[%d] = %d", m, n);
		break;
	case WS:
		m = *(stack--);
		n = *(stack--);
		stor[ n] = m;
		dprintf2( "WS %d -> store[%d]", m, n);
		break;

	// working on the control value table

	case WCVTP:
		m = *(stack--);
		n = *(stack--);
		i = m;
		cvt[ n] = i;
		dprintf3( "WCVTP %d -> cvt[%d] = %d", m, n, cvt[n]);
		break;
	case WCVTF:
		m = *(stack--);
		n = *(stack--);
		//### how does one scale a scalar with the ((xx,xy),(yx,yy)) matrix???
		cvt[ n] = ((m * ((xx + yy) >> 1) + 32) >> 6) << xxexp;
		dprintf3( "###WCVTF %d -> cvt[%d] = %d\n", m, n, cvt[n]);
		break;
	case RCVT:
		m = *stack;
		*stack = cvt[ m];
		dprintf2( "RCVT cvt[%d] = %d", m, cvt[m]);
		break;

	// accessing the graphics state

	case SVTCA0:
		dprintf0( "SVTCA0");
		gs.p_vector_x = gs.f_vector_x = 0;
		gs.p_vector_y = gs.f_vector_y = UNITY2D14;
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SVTCA1:
		dprintf0( "SVTCA1");
		gs.p_vector_x = gs.f_vector_x = UNITY2D14;
		gs.p_vector_y = gs.f_vector_y = 0;
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SPVTCA0:
		dprintf0( "SPVTCA0");
		gs.p_vector_x = 0;
		gs.p_vector_y = UNITY2D14;
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SPVTCA1:
		dprintf0( "SPVTCA1");
		gs.p_vector_x = UNITY2D14;
		gs.p_vector_y = 0;
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SFVTCA0:
		dprintf0( "SFVTCA0");
		gs.f_vector_x = 0;
		gs.f_vector_y = UNITY2D14;
		break;
	case SFVTCA1:
		dprintf0( "SFVTCA1");
		gs.f_vector_x = UNITY2D14;
		gs.f_vector_y = 0;
		break;
	case SDPVTL0:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "SDPVTL0 p[%d] p[%d]", m, n);
		oldLine2vector( gs.zp1[ n], gs.zp2[ m], gs.dp_vector_x, gs.dp_vector_y);
		newLine2vector( gs.zp1[ n], gs.zp2[ m], gs.p_vector_x, gs.p_vector_y);
		break;
	case SDPVTL1:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "SDPVTL1 p[%d] p[%d]", m, n);
		oldLine2vector( gs.zp1[ n], gs.zp2[ m], gs.dp_vector_y, gs.dp_vector_x);
		gs.dp_vector_x = -gs.dp_vector_x;
		newLine2vector( gs.zp1[ n], gs.zp2[ m], gs.p_vector_y, gs.p_vector_x);
		gs.p_vector_x = -gs.p_vector_x;
		break;
	case SPVTL0:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "SPVTL0 p[%d] p[%d]", m, n);
		newLine2vector( gs.zp1[ n], gs.zp2[ m], gs.p_vector_x, gs.p_vector_y);
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SPVTL1:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "SPVTL1 p[%d] p[%d]", m, n);
		newLine2vector( gs.zp1[ n], gs.zp2[ m], gs.p_vector_y, gs.p_vector_x);
		gs.p_vector_y = -gs.p_vector_y;
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SFVTL0:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "SFVTL0 p[%d] p[%d]", m, n);
		newLine2vector( gs.zp1[ n], gs.zp2[ m], gs.f_vector_x, gs.f_vector_y);
		break;
	case SFVTL1:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "SFVTL1 p[%d] p[%d]", m, n);
		newLine2vector( gs.zp1[ n], gs.zp2[ m], gs.f_vector_y, gs.f_vector_x);
		gs.f_vector_y = -gs.f_vector_y;
		break;
	case SFVTPV:
		dprintf0( "SFVTPV");
		gs.f_vector_x = gs.p_vector_x; 
		gs.f_vector_y = gs.p_vector_y; 
		break;
	case SPVFS:
		gs.p_vector_y = *(stack--);
		gs.p_vector_x = *(stack--);
		dprintf2( "#SPVFS = %d %d", gs.p_vector_x, gs.p_vector_y);
		gs.dp_vector_x = gs.p_vector_x;
		gs.dp_vector_y = gs.p_vector_y;
		break;
	case SFVFS:
		gs.f_vector_y = *(stack--);
		gs.f_vector_x = *(stack--);
		dprintf2( "#SFVFS = %d %d", gs.f_vector_x, gs.f_vector_y);
		break;
	case GPV:
		*(++stack) = gs.p_vector_x;
		*(++stack) = gs.p_vector_y;
		dprintf2( "GPV = %d %d", gs.p_vector_x, gs.p_vector_y);
		break;
	case GFV:
		*(++stack) = gs.f_vector_x;
		*(++stack) = gs.f_vector_y;
		dprintf2( "GFV = %d %d", gs.f_vector_x, gs.f_vector_y);
		break;
	case SRP0:
		gs.rp0 = *(stack--);
		dprintf1( "SRP0 p[%d]", gs.rp0);
		break;
	case SRP1:
		gs.rp1 = *(stack--);
		dprintf1( "SRP1 p[%d]", gs.rp1);
		break;
	case SRP2:
		gs.rp2 = *(stack--);
		dprintf1( "SRP2 p[%d]", gs.rp2);
		break;
	case SZP0:
		m = *(stack--);
		gs.zp0 = p[ m];
		dprintf1( "SZP0 %d", m);
		break;
	case SZP1:
		m = *(stack--);
		gs.zp1 = p[ m];
		dprintf1( "SZP1 %d", m);
		break;
	case SZP2:
		m = *(stack--);
		gs.zp2 = p[ m];
		dprintf1( "SZP2 %d", m);
		break;
	case SZPS:
		m = *(stack--);
		gs.zp2 = gs.zp1 = gs.zp0 = p[ m];
		dprintf1( "SZPS %d", m);
		if( m < 0 || m > 1) {
			dprintf0( "\nZPs out of range!!! ###");
		}
		break;
	case RTHG:
		dprintf0( "RTHG");
		gs.round_state	= ROUND_HALF;
		break;
	case RTG:
		dprintf0( "RTG");
		gs.round_state	= ROUND_GRID;
		break;
	case RTDG:
		dprintf0( "RTDG");
		gs.round_state	= ROUND_DOUBLE;
		break;
	case RDTG:
		dprintf0( "RDTG");
		gs.round_state	= ROUND_DOWN;
		break;
	case RUTG:
		dprintf0( "RUTG");
		gs.round_state	= ROUND_UP;
		break;
	case ROFF:
		dprintf0( "ROFF");
		gs.round_state	= ROUND_OFF;
		break;
	case SROUND:
		m = *(stack--);
		n = (m >> 6) & 3;
		i = (m >> 4) & 3;
		gs.round_state = ROUND_SUPER;
		gs.round_period = 0x20 << n;
		gs.round_phase = (gs.round_period >> 2) * i;
		m &= 0x0F;
		if( m)
			gs.round_thold = (gs.round_period >> 3) * (m - 4);
		else
			gs.round_thold = gs.round_period - 1;
		dprintf3( "SROUND %d %d %d", n, i, m);
		dprintf3( "-> period 0x%02X, thold 0x%02X, phase 0x%02X",
			gs.round_period, gs.round_thold, gs.round_phase);
		break;
	case S45ROUND:
		m = *(stack--);
		n = (m >> 6) & 3;
		i = (m >> 4) & 3;
		gs.round_state = ROUND_SUPER45;
		gs.round_period = 1444 >> (7 - n);
		gs.round_phase = (gs.round_period * i) >> 2;
		m &= 0x0F;
		if( m)
			gs.round_thold = (gs.round_period * (m - 4)) >> 3;
		else
			gs.round_thold = gs.round_period - 1;
		dprintf3( "SROUND45 %d %d %d", n, i, m);
		dprintf3( "-> period 0x%02X, thold 0x%02X, phase 0x%02X",
			gs.round_period, gs.round_thold, gs.round_phase);
		break;
	case SLOOP:
		gs.loop = *(stack--);
		dprintf1( "SLOOP %d",gs. loop);
		break;
	case SMD:
		gs.min_distance = *(stack--);
		dprintf1( "SMD %d", gs.min_distance);
		break;
	case INSTCTRL:
		gs.instr_control = *(stack--);
		m = *(stack--);
		dprintf2( "#INSTCTRL %d %d", gs.instr_control, m);
		if( gs.instr_control == 1)
			grid_fitting = !m;
		break;
	case SCANCTRL:
		gs.scan_control = *(stack--);
		dprintf1( "###SCANCTRL %d", gs.scan_control);
		//### what now???
		break;
	case SCANTYPE:
		n = *(stack--);
		dprintf1( "###SCANTYPE %d", n);
		//### what now?
		break;
	case SCVTCI:
		gs.cvt_cut_in = *(stack--);
		dprintf1( "SCVTCI %d", gs.cvt_cut_in);
		break;
	case SSWC:
		gs.swidth_cut_in = *(stack--);
		dprintf1( "SSWC %d", gs.swidth_cut_in);
		break;
	case FLIPON:
		gs.auto_flip = TRUE;
		dprintf0( "FLIPON");
		break;
	case FLIPOFF:
		gs.auto_flip = FALSE;
		dprintf0( "FLIPOFF");
		break;
	case SANGW:
		//angle_weight is obsolete!
		m = *(stack--);
		dprintf1( "SANGW %d is obsolete", m);
		break;
	case SDB:
		gs.delta_base = *(stack--);
		dprintf1( "SDB %d", gs.delta_base);
		break;
	case SDS:
		gs.delta_shift = *(stack--);
		dprintf1( "SDS %d", gs.delta_shift);
		break;

	// do some measurements

	case GC0:
		m = *stack;
		dprintf2( "GC0 p[%d][%d]\t", gs.zp2==p[1], m);
		n = absNewMeasure( gs.zp2[ m].xnow, gs.zp2[ m].ynow);
		*stack = n;
		dprintf1( "\t=> %d", n);
		break;
	case GC1:
		m = *stack;
		dprintf2( "GC1 p[%d][%d]\t", gs.zp2==p[1], m);
		n = absOldMeasure( gs.zp2[ m].xold, gs.zp2[ m].yold);
		*stack = n;
		dprintf1( "\t=> %d", n);
		break;
	case SCFS:
		// move point along freedom vector, so that
		// projection gets desired length
		m = *(stack--);
		n = *(stack--);
		dprintf3( "SCFS p[%d][%d] to %f", gs.zp2==p[1], n, m/FSHIFT);
		if( gs.zp2==p[1]) {
			pp = &p[ 1][ n];
			i = absNewMeasure( pp->xnow, pp->ynow);
			movePoint( *pp, m - i);
		} else {
			pp = &p[ 0][ n];
			pp->xold = pp->xnow = (m * gs.p_vector_x) >> 14;
			pp->yold = pp->ynow = (m * gs.p_vector_y) >> 14;
		}
		break;
	case MD0:
		m = *(stack--);
		n = *stack;
		dprintf2( "#MD0 p[%d][%d] ", gs.zp1==p[1], m);
		dprintf2( "- p[%d][%d]", gs.zp2==p[1], n);
		//###i = newMeasure( gs.zp1[ m], gs.zp2[ n]);
		i = newMeasure( gs.zp2[ n], gs.zp1[ m]);
		*stack = i;
		break;
	case MD1:
		m = *(stack--);
		n = *stack;
		dprintf2( "#MD1 p[%d][%d] ", gs.zp1==p[1], m);
		dprintf2( "- p[%d][%d]", gs.zp2==p[1], n);
		//###i = oldMeasure( gs.zp1[ m], gs.zp2[ n]);
		i = oldMeasure( gs.zp2[ n], gs.zp1[ m]);
		*stack = i;
		break;
	case MPPEM:
		dprintf0( "MPPEM\t");
		m = absNewMeasure( mppemx, mppemy);
		*(++stack) = m;
		dprintf1( "\t => mppem = %d", m);
		break;
	case MPS:
		m = pointSize;
		*(++stack) = m;
		dprintf1( "MPS %d", m);
		break;

	// outline manipulation

	case FLIPPT:
		for( m = gs.loop; --m >= 0;) {
			n = *(stack--);
			dprintf3( "FLIPPT * %d p[%d][%d]", m, gs.zp0==p[1], n);
			gs.zp0[ n].flags ^= ON_CURVE;
		}
		gs.loop = 1;
		break;
	case FLIPRGON:
		m = *(stack--);
		n = *(stack--);
		dprintf3( "FLIPRGON p[%d][ %d .. %d]", gs.zp0==p[1], n, m);
		pp = &gs.zp0[ n];
		while( n++ <= m)
			(pp++)->flags |= ON_CURVE;
		break;
	case FLIPRGOFF:
		m = *(stack--);
		n = *(stack--);
		dprintf3( "FLIPRGOFF p[%d][ %d .. %d]", gs.zp0==p[1], n, m);
		pp = &gs.zp0[ n];
		while( n++ <= m)
			(pp++)->flags &= ~ON_CURVE;
		break;
	case SHP0:
	case SHP1:
		for( m = gs.loop; --m >= 0;) {
			pp = (opc & 1) ? &gs.zp0[ gs.rp1] : &gs.zp1[ gs.rp2];
			i = *(stack--);
			dprintf3( "SHP * %d p[%d], rp = p[%d]\n", m, i, pp-p[1]);
			n = absNewMeasure( pp->xnow - pp->xold, pp->ynow - pp->yold);
			dprintf1( "rp moved by %f", n/FSHIFT);
			movePoint( gs.zp2[ i], n);
		}
		gs.loop = 1;
		break;
	case SHC0:
	case SHC1:
		m = *(stack--);
		pp = (opc & 1) ? &gs.zp0[ gs.rp1] : &gs.zp1[ gs.rp2];
		n = absNewMeasure( pp->xnow - pp->xold, pp->ynow - pp->yold);
		i = (m <= 0) ? 0 : endPoints[ m-1];
		m = (gs.zp2==p[ 0]) ? nPointsZ0 : endPoints[ m];
		for( ; i <= m; ++i) {
			dprintf3( "SHC%d p[%d] by %f\n", opc&1, i, n/FSHIFT);
			movePoint( gs.zp2[ i], n);
		}
		break;
	case SHZ0:
	case SHZ1:
		m = *(stack--);
		pp = (opc & 1) ? &gs.zp0[ gs.rp1] : &gs.zp1[ gs.rp2];
		n = absNewMeasure( pp->xnow - pp->xold, pp->ynow - pp->yold);
		i = (m <= 0) ? nPointsZ0 : nPointsZ1;
		while( --i >= 0) {
			if( ((opc&1)? gs.rp1: gs.rp2) == i) continue;
			dprintf3( "SHZ%d p[%d] by %f", opc&1, i, n/FSHIFT);
			dprintf1( ", rp = p[%d]", (opc&1) ? gs.rp1 : gs.rp2);
			movePoint( p[m][i], n);
		}
		break;
	case SHPIX:
		m = *(stack--);
		for( n = gs.loop; --n >= 0;) {
			i = *(stack--);
			dprintf3( "###SHPIX * %d p[%d][%d] ", n, gs.zp2==p[1], i);
			dprintf1( "by %f", m/FSHIFT);
			pp = &gs.zp2[ i];
			dprintf2( "\n%d %d ->", pp->xnow, pp->ynow);
			pp->xnow += (m * gs.f_vector_x) >> 14;
			pp->ynow += (m * gs.f_vector_y) >> 14;
			dprintf2( "\t%d %d", pp->xnow, pp->ynow);
		}
		gs.loop = 1;
		break;
	case MSIRP0:
	case MSIRP1:
		m = *(stack--);
		n = *(stack--);
		gs.rp2 = n;
		gs.rp1 = gs.rp0;
		dprintf3( "MSIRP%d p[%d][%d] ", opc&1, gs.zp1==p[1], n);
		dprintf3( "to %f, rp = p[%d][%d]", m/FSHIFT, gs.zp0==p[1], gs.rp0);
		if( gs.zp1 == p[1]) {
			i = newMeasure( p[1][n], gs.zp0[ gs.rp0]);
			movePoint( p[1][n], m - i);
		} else {	// magic in the twilight zone
			pp = &p[ 0][ n];
			pp->xnow = pp->xold =
				((m * gs.p_vector_x) >> 14) + gs.zp0[ gs.rp0].xnow;
			pp->ynow = pp->yold =
				((m * gs.p_vector_y) >> 14) + gs.zp0[ gs.rp0].ynow;
		}
		if( opc & 1)
			gs.rp0 = n;
		break;
	case MDAP0:
	case MDAP1:
		gs.rp0 = gs.rp1 = m = *(stack--);
		dprintf2( "MDAP%d p[%d]", opc&1, m);
		pp = &gs.zp0[ m];
		dprintf2( "\nxy %d %d", pp->xnow, pp->ynow);
		if( gs.f_vector_x)
				pp->flags |= X_TOUCHED;
		if( gs.f_vector_y)
				pp->flags |= Y_TOUCHED;
		if( opc & 1) {
			if( gs.f_vector_x)
				pp->xnow = round( pp->xnow);
			if( gs.f_vector_y)
				pp->ynow = round( pp->ynow);
			dprintf2( "\t-> %d %d", pp->xnow, pp->ynow);
		}
		break;
	case MIAP0:
	case MIAP1:
		m = *(stack--);
		n = *(stack--);
		gs.rp0 = gs.rp1 = n;
		dprintf3( "MIAP%d p[%d][%d] ", opc&1, gs.zp0==p[1], n);
		dprintf1( "to cvt[%d] = ", m);
		m = cvt[ m];
		dprintf1( "%f", m/FSHIFT);
		if( gs.zp0 == p[0]) {	// magic in the twilight zone
			i = m;
			pp = &p[0][n];
			pp->xold = pp->xnow = (m * gs.p_vector_x) >> 14;
			pp->yold = pp->ynow = (m * gs.p_vector_y) >> 14;
			if( opc & 0x01) {
				pp->xnow = round( pp->xnow);
				pp->ynow = round( pp->ynow);
			}
		} else {
			pp = &p[1][n];
			i = absNewMeasure( pp->xnow, pp->ynow);
			if( opc & 0x01) {
				if( ((m>i) ? m-i : i-m) > gs.cvt_cut_in)
					m = i;
				m = round( m);
			}
			dprintf1( "\nabsdist = %f", i/FSHIFT);
			movePoint( gs.zp0[ n], m - i);
		}
		break;
	case MDRP00: case MDRP01:
	case MDRP02: case MDRP03:
	case MDRP04: case MDRP05:
	case MDRP06: case MDRP07:
	case MDRP08: case MDRP09:
	case MDRP0A: case MDRP0B:
	case MDRP0C: case MDRP0D:
	case MDRP0E: case MDRP0F:
	case MDRP10: case MDRP11:
	case MDRP12: case MDRP13:
	case MDRP14: case MDRP15:
	case MDRP16: case MDRP17:
	case MDRP18: case MDRP19:
	case MDRP1A: case MDRP1B:
	case MDRP1C: case MDRP1D:
	case MDRP1E: case MDRP1F:
		gs.rp2 = m = *(stack--);
		gs.rp1 = gs.rp0;
		dprintf3( "#MDRP%02X p[%d], rp = p[%d]", opc&15, m, gs.rp0);
		n = oldMeasure( gs.zp1[ m], gs.zp0[ gs.rp0]);
		i = newMeasure( gs.zp1[ m], gs.zp0[ gs.rp0]);
		dprintf2( "\nwanteddist = %f, currentdist = %f", n/FSHIFT, i/FSHIFT);
		dprintf2( "\np = (%d %d)", gs.zp1[ m].xnow, gs.zp1[ m].ynow);
		dprintf2( " - rp = (%d %d)", gs.zp0[ gs.rp0].xnow, gs.zp0[ gs.rp0].ynow);

		if( ((n >= 0) ? +n : -n) < gs.swidth_cut_in)
			n = (n >= 0) ? +gs.swidth_value : -gs.swidth_value;
		if( opc & 0x10)
			gs.rp0 = m;
		if( opc & 0x04)
			n = round( n);
		if( opc & 0x08)
			if( n >= 0) {
				if( n < +gs.min_distance)
					n = +gs.min_distance;
			} else {
				if( n > -gs.min_distance)
					n = -gs.min_distance;
			}
		//### ignore black/gray/white for now
		movePoint( gs.zp1[ m], n - i);
		break;
	case MIRP00: case MIRP01:
	case MIRP02: case MIRP03:
	case MIRP04: case MIRP05:
	case MIRP06: case MIRP07:
	case MIRP08: case MIRP09:
	case MIRP0A: case MIRP0B:
	case MIRP0C: case MIRP0D:
	case MIRP0E: case MIRP0F:
	case MIRP10: case MIRP11:
	case MIRP12: case MIRP13:
	case MIRP14: case MIRP15:
	case MIRP16: case MIRP17:
	case MIRP18: case MIRP19:
	case MIRP1A: case MIRP1B:
	case MIRP1C: case MIRP1D:
	case MIRP1E: case MIRP1F:
		m = *(stack--);
		n = *(stack--);
		gs.rp2 = n;
		gs.rp1 = gs.rp0;
		dprintf3( "#MIRP%02X p[%d] with cvt[%d]", opc&15, n, m);
		m = cvt[ m];
		dprintf2( " = %f, rp = p[%d]", m/FSHIFT, gs.rp0);

		if( ((m >= 0)? +m : -m) < +gs.swidth_cut_in)
			m = (m >= 0) ? +gs.swidth_value : -gs.swidth_value;

		i = newMeasure( gs.zp1[ n], gs.zp0[ gs.rp0]);
		j = oldMeasure( gs.zp1[ n], gs.zp0[ gs.rp0]);

		if( (j^m) < 0 && gs.auto_flip) {
			m = -m;
			dprintf1( "\nautoflip m = %f", m/FSHIFT);
		}
		if( opc & 0x10)
			gs.rp0 = n;
		if( opc & 0x04) {
			if( ((m>j) ? m-j : j-m) >= +gs.cvt_cut_in)
				m = j;
			m = round( m);
		}
		if( opc & 0x08) {
			if( j >= 0) {
				if( m < +gs.min_distance)
					m = +gs.min_distance;
 			} else {
				if( m > -gs.min_distance)
					m = -gs.min_distance;
			}
		}
		//### ignore device dependend rounding for now
		movePoint( gs.zp1[ n], m - i);
		break;
	case ALIGNRP:
		for( m = gs.loop; --m >= 0;) {
			n = *(stack--);
			dprintf3( "ALIGNRP * %d p[%d], rp0 = p[%d]", m, n, gs.rp0);
			i = newMeasure( gs.zp1[ n], gs.zp0[ gs.rp0]);
			movePoint( gs.zp1[ n], -i);
		}
		gs.loop = 1;
		break;
	case ALIGNPTS:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "ALIGNPTS %d %d", m, n);
		i = newMeasure( gs.zp1[ m], gs.zp0[ n]) >> 1;
		movePoint( gs.zp0[ n], +i);
		movePoint( gs.zp1[ m], -i);
		break;
	case ISECT:
		{
		point* pp1 = &gs.zp1[ *(stack--)];
		point* pp2 = &gs.zp1[ *(stack--)];
		point* pp3 = &gs.zp0[ *(stack--)];
		point* pp4 = &gs.zp0[ *(stack--)];
		m = *(stack--);

		dprintf1( "ISECT p[%d] ", m);
		dprintf2( "between p[%d]-p[%d] ", pp1-gs.zp1, pp2-gs.zp1);
		dprintf2( "and p[%d]-p[%d] ", pp3-gs.zp0, pp4-gs.zp0);

		int f1 = (pp1->xnow - pp3->xnow) * (pp4->ynow - pp3->ynow)
			-(pp1->ynow - pp3->ynow) * (pp4->xnow - pp3->xnow);
		int f2 = (pp2->ynow - pp1->ynow) * (pp4->xnow - pp3->xnow)
			-(pp2->xnow - pp1->xnow) * (pp4->ynow - pp3->ynow);

		pp3 = &gs.zp2[ m];
		pp3->flags |= X_TOUCHED | Y_TOUCHED;
		if( f2 == 0 ) { // parallel !!!
			pp3->xnow = (pp2->xnow + pp1->xnow + 1) >> 1;
			pp3->ynow = (pp2->ynow + pp1->ynow + 1) >> 1;
			dprintf0( "are parallel!\n");
		} else {
			float f = (float)f1 / f2;
			pp3->xnow = (int)(f * (pp2->xnow - pp1->xnow)) + pp1->xnow;
			pp3->ynow = (int)(f * (pp2->ynow - pp1->ynow)) + pp1->ynow;
		}

		dprintf2( "\n-> %d %d", pp3->xnow, pp3->ynow);
		}
		break;
	case AA:
		stack--;
		dprintf0( "AA is obsolete and not supported!");
		break;
	case IP:
		for( m = gs.loop; --m >= 0;) {
			n = *(stack--);
			dprintf2( "IP * %d p[%d] ", m, n);
			dprintf2( "between p[%d][%d] ", gs.zp1==p[1], gs.rp2);
			dprintf2( "and p[%d][%d]", gs.zp0==p[1], gs.rp1);
			interpolate( gs.zp2[ n],
				gs.zp1[ gs.rp2], gs.zp0[ gs.rp1]);
			dprintf0( "\n");
		}
		gs.loop = 1;
		break;
	case UTP:
		m = *(stack--);
		gs.zp0[ m].flags &= ~(X_TOUCHED | Y_TOUCHED);
		dprintf1( "UTP p[%d]", m);
		break;
	case IUP0:
		for( m = n = 0; m < nEndPoints; ++m) {
			i = endPoints[ m];
			dprintf2( "IUP0 p[%d .. %d]", n, i);
			doIUP0( n, i);
			n = i + 1;
		}
		break;
	case IUP1:
		for( m = n = 0; m < nEndPoints; ++m) {
			i = endPoints[ m];
			dprintf2( "IUP1 p[%d .. %d]", n, i);
			doIUP1( n, i);
			n = i + 1;
		}
		break;
	case DELTAP1:
		n = 0;
		goto deltap_label;
	case DELTAP2:
		n = -16;
		goto deltap_label;
	case DELTAP3:
		n = -32;
deltap_label:
		m = *(stack--);
		dprintf2( "###DELTAP%d * %d", (-n >> 4) + 1, m);
		n += mppem - gs.delta_base;
		if( n < 0 || n > 15) {
			dprintf1( "\n=> skipping %d exceptions", m);
			stack -= m << 1;
			break;
		}
		n <<= 4;
		while( --m >= 0) {
			int pno = *(stack--);
			int arg = *(stack--);
			dprintf2( "\np[%d] arg %04X", pno, arg);
			dprintf2( "\targ.n=%d, n=%d", arg>>4, n>>4);
			if( n > (arg & 0xf0))
				break;
			if( n == (arg & 0xf0)) {
				arg = (arg & 15) - 8;
				if( arg >= 0) ++arg;
				arg <<= (SHIFT - gs.delta_shift);
				dprintf1( "\tmoving by %f", arg/FSHIFT);
				movePoint( gs.zp0[ pno], arg);
				//###gs.zp0[ pno].xnow += (arg * gs.f_vector_x) >> 14;
				//###gs.zp0[ pno].ynow += (arg * gs.f_vector_y) >> 14;
			}
		}
		while( --m >= 0) {
			int pno = *(stack--);
			int arg = *(stack--);
			dprintf2( "\n(p[%d] arg %04X", pno, arg);
			dprintf2( "\targ.n=%d, n=%d)", arg>>4, n>>4);
		};
		break;
	case DELTAC1:
		n = 0;
		goto deltac_label;
	case DELTAC2:
		n = -16;
		goto deltac_label;
	case DELTAC3:
		n = -32;
deltac_label:
		m = *(stack--);
		dprintf2( "###DELTAC%d * %d", (-n >> 4) + 1, m);
		n += mppem - gs.delta_base;
		if( n < 0 || n > 15) {
			stack -= m << 1;
			break;
		}
		n <<= 4;
		while( --m >= 0) {
			int cno = *(stack--);
			int arg = *(stack--);
			dprintf3( "\ncvt[%d] arg = %04X, n = %d", cno, arg, n>>4);
			if( n > (arg & 0xf0))
				break;
			if( n == (arg & 0xf0)) {
				arg = (arg & 15) - 8;
				if( arg >= 0) ++arg;
				arg <<= SHIFT - gs.delta_shift;
				dprintf2( "\tmoved by %f,\t%d ", arg/FSHIFT, cvt[cno]);
				cvt[ cno] += arg;
				dprintf1( "-> %d", cvt[cno]);
			}
		}
		while( --m >= 0) {
			int cno = *(stack--);
			int arg = *(stack--);
			dprintf2( "\n(cvt[%d] arg %04X", cno, arg);
			dprintf2( "\targ.n=%d, n=%d)", arg>>4, n>>4);
		};
		break;

	// stack manipulation

	case DUP:
		m = *stack;
		*(++stack) = m;
		dprintf1( "DUP = %d", m);
		break;
	case POP:
		dprintf0( "POP");
		--stack;
		break;
	case CLEAR:
		dprintf0( "CLEAR");
		stack = stackbase;
		break;
	case SWAP:
		dprintf0( "SWAP");
		m = *stack;
		*stack = *(stack-1);
		*(stack-1) = m;
		break;
	case DEPTH:
		m = stack - stackbase;
		*(++stack) = m;
		dprintf1( "DEPTH = %d", m);
		break;
	case CINDEX:
		m = *stack;
		n = *(stack - m);
		*stack = n;
		dprintf2( "CINDEX %d = %d", m, n);
		break;
	case MINDEX:
		m = *stack;
		stack -= m;
		n = *stack;
		dprintf2( "MINDEX %d = %d", m, n);
		for( ; --m > 0; ++stack)
			*stack = stack[1];
		*stack = n;
		break;
	case ROLL:
		m = *(stack - 0);
		n = *(stack - 1);
		i = *(stack - 2);
		*(stack - 0) = i;
		*(stack - 1) = m;
		*(stack - 2) = n;
		dprintf3( "ROLL %d %d %d", m, n, i);
		dprintf3( " => %d %d %d", i, m, n);
		break;

	// control flow

	case IF:
		m = *(stack--);
		dprintf1( "IF %d", m);
		if( !m)
			skipHints( f, IF);
		break;
	case ELSE:
		// if we hit ELSE we didn't skip -> skip from here
		dprintf0( "ELSE");
		skipHints( f, ELSE);
		break;
	case EIF:
		dprintf0( "EIF");
		break;
	case JROT:
		n = *(stack--);
		dprintf1( "JROT %d -> ", n);
		if( n)
			goto jump_relative;
		dprintf0( "not taken");
		--stack;
		break;
	case JROF:
		n = *(stack--);
		dprintf1( "JROF %d -> ", n);
		if( !n)
			goto jump_relative;
		dprintf0( "not taken");
		--stack;
		break;
	case JMPR:
jump_relative:
		m = *(stack--);
		dprintf1( "JMPR %d", m);
		f->seekRelative( m-1);
		break;
	case LT:
		m = *(stack--);
		n = *stack;
		i = (n < m);
		*stack = i;
		dprintf3( "LT %d %d = %d", m, n, i);
		break;
	case LTEQ:
		m = *(stack--);
		n = *stack;
		i = (n <= m);
		*stack = i;
		dprintf3( "LTEQ %d %d = %d", m, n, i);
		break;
	case GT:
		m = *(stack--);
		n = *stack;
		i = (n > m);
		*stack = i;
		dprintf3( "GT %d %d = %d", m, n, i);
		break;
	case GTEQ:
		m = *(stack--);
		n = *stack;
		i = (n >= m);
		*stack = i;
		dprintf3( "GTEQ %d %d = %d", m, n, i);
		break;
	case EQ:
		m = *(stack--);
		n = *stack;
		i = (m == n);
		*stack = i;
		dprintf3( "EQ %d %d = %d", m, n, i);
		break;
	case NEQ:
		m = *(stack--);
		n = *stack;
		i = (m != n);
		*stack = i;
		dprintf3( "NEQ %d %d = %d", m, n, i);
		break;
	case ODD:
		m = *stack;
		n = (round(m) >> SHIFT) & 1;
		*stack = n;
		dprintf2( "ODD %d = %d", m, n);
		break;
	case EVEN:
		m = *stack;
		n = ((~round(m)) >> SHIFT) & 1;
		*stack = n;
		dprintf2( "EVEN %d = %d", m, n);
		break;
	case AND:
		m = *(stack--);
		n = *stack;
		i = n && m;
		*stack = i;
		dprintf3( "AND %d %d = %d", m, n, i);
		break;
	case OR:
		m = *(stack--);
		n = *stack;
		i = n || m;
		*stack = i;
		dprintf3( "OR %d %d = %d", m, n, i);
		break;
	case NOT:
		m = *stack;
		n = !m;
		*stack = n;
		dprintf2( "NOT %d = %d", m, n);
		break;
	case ADD:
		m = *(stack--);
		n = *stack;
		i = m + n;
		*stack = i;
		dprintf3( "ADD %d %d = %d", m, n, i);
		break;
	case SUB:
		m = *(stack--);
		n = *stack;
		i = n - m;
		*stack = i;
		dprintf3( "SUB %d %d = %d", m, n, i);
		break;
	case DIV:
		m = *(stack--);
		n = *stack;
		if( m) i = (n << SHIFT) / m;
		else i = (n>=0) ? 0x7fffff : -0x7ffffff;
		*stack = i;
		dprintf3( "DIV %d %d = %d", m, n, i);
		break;
	case MUL:
		m = *(stack--);
		n = *stack;
		i = (m * n) >> SHIFT;
		*stack = i;
		dprintf3( "MUL %d %d = %d", m, n, i);
		break;
	case ABS:
		m = *stack;
		n = (m >= 0) ? m : -m;
		*stack = n;
		dprintf2( "ABS %d = %d", m, n);
		break;
	case NEG:
		m = *stack;
		n = -m;
		*stack = n;
		dprintf2( "NEG %d = %d", m, n);
		break;
	case FLOOR:
		m = *stack;
		n = m & -64;
		*stack = n;
		dprintf2( "FLOOR %d = %d", m, n);
		break;
	case CEILING:
		m = *stack;
		n = (m + 63) & -64;
		*stack = n;
		dprintf2( "CEILING %d = %d", m, n);
		break;
	case MAX:
		m = *(stack--);
		n = *stack;
		i = (m >= n) ? m : n;
		*stack = i;
		dprintf3( "MAX %d %d = %d", m, n, i);
		break;
	case MIN:
		m = *(stack--);
		n = *stack;
		i = (m <= n) ? m : n;
		*stack = i;
		dprintf3( "MIN %d %d = %d", m, n, i);
		break;
	case ROUND00: case ROUND01:
	case ROUND02: case ROUND03:
		// 00: gray, 01: black, 02: white, 03: ???
		m = *stack;
		//### no engine compensation for now
		n = round( m);
		*stack = n;
		dprintf3( "#ROUND%02X %d = %d", opc&3, m, n);
		break;
	case NROUND00: case NROUND01:
	case NROUND02: case NROUND03:
		// 00: gray, 01: black, 02: white, 03: ???
		m = *stack;
		//### no engine compensation for now
		n = round( m);
		*stack = n;
		dprintf3( "#NROUND%02X %d = %d", opc&3, m, n);
		break;
	case FDEF:
		m = *(stack--);
		dprintf1( "FDEF %d", m);
		fdefs[ m].f	= f;
		fdefs[ m].offset = f->tellRel();
		skipHints( f, FDEF);
		fdefs[ m].length = f->tellRel() - fdefs[ m].offset;
		break;
	case ENDF:
		dprintf0( "ENDF\n");
		return;
		break;
	case IDEF:
		m = *(stack--);
		dprintf1( "IDEF %d", m);
		idefs[ m].f	= f;
		idefs[ m].offset = f->tellRel();
		skipHints( f, IDEF);
		idefs[ m].length = f->tellRel() - idefs[ m].offset;
		break;
	case CALL:
		m = *(stack--);
		dprintf1( "CALL %d\n", m);
		n = f->tellRel();
		{
			FDefs* fd = &fdefs[m];
			execHints( fd->f, fd->offset, fd->length);
		}
		f->seekAbsolute( n);
		break;
	case LOOPCALL:
		m = *(stack--);
		n = *(stack--);
		dprintf2( "LOOPCALL %d * %d\n", m, n);
		i = f->tellRel();
		{
			FDefs* fd = &fdefs[m];
			while( --n >= 0)
				execHints( fd->f, fd->offset, fd->length);
		}
		f->seekAbsolute( i);
		break;
	case DBG:
		dprintf0( "DBG not implemented");
		break;
	case GETINFO:
		m = *stack;
		n = 0;
		if( m & SCALER_VERSION)
			n = WIN_SCALER;
		if( m & GLYPH_ROTATED)
			if( xy | yx)
				n |= IS_ROTATED;
		if( m & GLYPH_STRETCHED)
			if( xx != yy)
				n |= IS_STRETCHED;
		*stack = n;
		dprintf2( "GETINFO %d = 0x%03X", m, n);
		break;
	default:
		dprintf1( "unknown instruction 0x%02X\n", opc);
		break;
	}
}

void Rasterizer::execHints( RandomAccessFile* f, int offset, int length)
{
	if( (unsigned)length > 10000) { //### catch the culprit!
		printf( "execHints.length = %d\n", length);
		exit( -1);
	}

	f->seekAbsolute( offset);

	length += offset;
	while( f->tellRel() < length) {
		int opc = f->readUByte();
		dprintf1( "\n::%05X\t", f->tellAbs() - 1);
		execOpcode( opc, f);
	}
	dprintf0( "\n\n");
}

void Rasterizer::skipHints( RandomAccessFile* f, int opc)
{
	register int m, depth = 1;
	dprintf0( "\nskipping...");
	do {
		switch( opc = f->readUByte()) {
		case NPUSHB:
			opc = f->readUByte() + PUSHB00 - 1;
			// fall through
		case PUSHB00: case PUSHB01:
		case PUSHB02: case PUSHB03:
		case PUSHB04: case PUSHB05:
		case PUSHB06: case PUSHB07:
			m = opc - PUSHB00 + 1;
			f->seekRelative( m);
			break;
		case NPUSHW:
			opc = f->readUByte() + PUSHW00 - 1;
			// fall through
		case PUSHW00: case PUSHW01:
		case PUSHW02: case PUSHW03:
		case PUSHW04: case PUSHW05:
		case PUSHW06: case PUSHW07:
			m = opc - PUSHW00 + 1;
			f->seekRelative( m << 1);
			break;
		case IF:
		case FDEF:
		case IDEF:
			++depth;
			break;
		case ELSE:
			if( depth == 1)
				return;
			break;
		case EIF:
		case ENDF:
			--depth;
			break;
		}
	} while( depth > 0);
}

void Rasterizer::movePoint( point& pp, int len11D6)
{
	int unit = absNewMeasure( gs.f_vector_x, gs.f_vector_y);
	if( !unit)
		return;

	dprintf3( "\nmovePoint p[%d] by %f units -> %f points", &pp-p[1],
		len11D6/FSHIFT, len11D6*256.0/unit);
	dprintf2( "\nxy %d %d", pp.xnow, pp.ynow); 

	pp.xnow += (len11D6 * gs.f_vector_x) / unit;
	pp.ynow += (len11D6 * gs.f_vector_y) / unit;

	if( gs.f_vector_x)
		pp.flags |= X_TOUCHED;
	if( gs.f_vector_y)
		pp.flags |= Y_TOUCHED;

	dprintf2( "\t-> %d %d\n", pp.xnow, pp.ynow); 
}

void Rasterizer::interpolate( point& pp, point& pRef1, point& pRef2)
{
	// interpolate pp, so that
	//
	//   dist(pp, pRef1)        dist(pp, pRef2)
	// ------------------- == -------------------
	//  dist(pp', pRef1')      dist(pp', pRef2')

	int dold21 = oldMeasure( pRef2, pRef1);
	int doldp1 = oldMeasure( pp, pRef1);

	int dist;

	if( (dold21 ^ doldp1) < 0)
		dist = absNewMeasure( pRef1.xnow - pRef1.xold,
			pRef1.ynow - pRef1.yold)
		     - absNewMeasure( pp.xnow - pp.xold,
			pp.ynow - pp.yold);
	else if( (dold21 ^ (dold21 - doldp1)) < 0)
		dist = absNewMeasure( pRef2.xnow - pRef2.xold,
			pRef2.ynow - pRef2.yold)
		     - absNewMeasure( pp.xnow - pp.xold,
			pp.ynow - pp.yold);
	else if( dold21 == 0)
		return;
	else {
		int dnew21 = newMeasure( pRef2, pRef1)<<1;
		int dnewp1 = newMeasure( pp, pRef1);
		//### strange, isn't it? Does anybody have a better formula?
		//###dist = (int)((((S64)doldp1 * dnew21) / dold21)>>1) - dnewp1;
		dist = ((int)(((float)doldp1 * dnew21) / dold21)>>1) - dnewp1;
	}

	dprintf1( "\nmove by %f", dist/FSHIFT);
	movePoint( pp, dist);
}

void Rasterizer::iup0( point& pp, point& pRef1, point& pRef2)
{
	int dold21 = pRef2.yold - pRef1.yold;
	int doldp1 = pp.yold - pRef1.yold;

	dprintf3( "\np[%d] between p[%d] and p[%d]", &pp-p[1], &pRef1-p[1], &pRef2-p[1]);
	dprintf2( "\nd21o dp1o %f %f", dold21/FSHIFT, doldp1/FSHIFT);

	dprintf2( "\tchanging y: %d %d", pp.xnow, pp.ynow);

	if( (dold21 ^ doldp1) < 0 || dold21 == 0)
		pp.ynow = pp.yold + pRef1.ynow - pRef1.yold;
	else if( (dold21 ^ (dold21 - doldp1)) < 0)
		pp.ynow = pp.yold + pRef2.ynow - pRef2.yold;
	else {
		double dnew21 = pRef2.ynow - pRef1.ynow;
		dprintf1( "\nd21n %8.3f", dnew21/FSHIFT);
		pp.ynow = (int)((doldp1 * dnew21) / dold21 + pRef1.ynow);
	}

	dprintf2( " -> %d %d\n", pp.xnow, pp.ynow);
}

void Rasterizer::iup1( point& pp, point& pRef1, point& pRef2)
{
	int dold21 = pRef2.xold - pRef1.xold;
	int doldp1 = pp.xold - pRef1.xold;
	
	dprintf3( "\np[%d] between p[%d] and p[%d]",
		&pp-p[1], &pRef1-p[1], &pRef2-p[1]);
	dprintf2( "\nd21o dp1o %f %f", dold21/FSHIFT, doldp1/FSHIFT);

	dprintf2( "\tchanging x: %d %d", pp.xnow, pp.ynow);

	if( (dold21 ^ doldp1) < 0 || dold21 == 0)
		pp.xnow = pp.xold + pRef1.xnow - pRef1.xold;
	else if( (dold21 ^ (dold21 - doldp1)) < 0)
		pp.xnow = pp.xold + pRef2.xnow - pRef2.xold;
	else {
		double dnew21 = pRef2.xnow - pRef1.xnow;
		dprintf1( "\nd21n %8.3f", dnew21/FSHIFT);
		pp.xnow = (int)((doldp1 * dnew21) / dold21 + pRef1.xnow);
	}

	dprintf2( " -> %d %d\n", pp.xnow, pp.ynow);
}

void Rasterizer::doIUP0( int first, int last)
{
	int p0;
	for( p0 = first; p0 <= last; ++p0)
		if( p[1][p0].flags & Y_TOUCHED)
			break;
	int i, j;
	for( i = j = p0; i <= last; i = j) {
		while( ++j <= last)
			if( p[1][j].flags & Y_TOUCHED)
				break;
		if( j > last)
			break;
		for( int k = i; ++k < j;)
			iup0( p[1][k], p[1][i], p[1][j]);
	}
	if( i > last)
		return;
	int k;	// MS VC++ uses obsolete scoping
	for( k = i; ++k <= last;)
		iup0( p[1][k], p[1][i], p[1][p0]);
	for( k = first; k < p0; ++k)
		iup0( p[1][k], p[1][i], p[1][p0]);
}

void Rasterizer::doIUP1( int first, int last)
{
	int p0;
	for( p0 = first; p0 <= last; ++p0)
		if( p[1][p0].flags & X_TOUCHED)
			break;
	int i = p0;
	int j = p0;
	while( i <= last) {
		while( ++j <= last)
			if( p[1][j].flags & X_TOUCHED)
				break;
		if( j > last)
			break;
		for( int k = i; ++k < j;)
			iup1( p[1][k], p[1][i], p[1][j]);
		i = j;
	}
	if( i > last)
		return;
	int k;	// MS VC++ uses obsolete scoping
	for( k = i; ++k <= last;)
		iup1( p[1][k], p[1][i], p[1][p0]);
	for( k = first; k < p0; ++k)
		iup1( p[1][k], p[1][i], p[1][p0]);
}

