// scaler
// Copyright 1997 Herbert Duerr

#include "xfstt.h"
#include <limits.h>
#ifdef WIN32
	#include <windows.h>
	#include <memory.h>
#else
	#include <sys/mman.h>
	//#include <string.h>
	#include <linux/string.h>
#endif

Rasterizer::Rasterizer( int _grid_fitting, int _anti_aliasing)
:
	stackbase( 0),
	ttFont( 0),
	status( INVALID_FONT),
	grid_fitting( _grid_fitting),
	anti_aliasing( _anti_aliasing)
{
	endBMP = 0;
}


Rasterizer::~Rasterizer()
{
	endInterpreter();
}


void Rasterizer::useTTFont( TTFont* _ttFont, int _flags)
{
	flags = _flags;

	// set scaling defaults = 16points
	xx = yy = (16 * 96 + 36) / 72;
	xy = yx = 0;

	if( ttFont != _ttFont) {
		ttFont = _ttFont;
		if( ttFont && ttFont->checkFont()) {
			initInterpreter();
			status = NOT_READY;
		} else
			status = INVALID_FONT;
	}
}

// prepare scaling of outlines

void Rasterizer::setTransformation( int _xx, int _xy, int _yx, int _yy,
	int xres, int yres)
{
	dprintf2( "_xx = %d,\t_xy = %d\n", _xx, _xy);
	dprintf2( "_yx = %d,\t_yy = %d\n", _xy, _yy);

	format = 0;

	pointSize = (_xx + _yy) / 2;	//### quick hack

	xx = (_xx * xres + 36) / 72;
	xy = (_xy * xres + 36) / 72;
	yx = (_yx * yres + 36) / 72;
	yy = (_yy * yres + 36) / 72;

	mppemx = xx;
	mppemy = yy;

	// 0177 would make more sense than 177, wouldn't it?
	// but since the reference implementation does it, we should do it too
	for( xxexp = 0; xx>177 || xx<-177 || yy>177 || yy<-177; ++xxexp) {
		xx >>= 1; xy >>= 1;
		yx >>= 1; yy >>= 1;
	}

	if( ttFont)
		applyTransformation();
	else
		status = NOT_READY;
}


// only needed when emUnits != 2048, but since we don't know this until
// we know the TTFont we have to separate it from setTransformation()

void Rasterizer::applyTransformation()
{
	int emUnits = ttFont->getEmUnits();

	for(; emUnits > 2048; ++xxexp)
		emUnits >>= 1;
	for(; emUnits < 2048 && xxexp > 0; --xxexp) 
		emUnits <<= 1;
	if( emUnits != 2048) {
		xx = (xx << 11) / emUnits;
		xy = (xy << 11) / emUnits;
		yx = (yx << 11) / emUnits;
		yy = (yy << 11) / emUnits;
	}

	mppem = (mppemx + mppemy) >> 1;

	dprintf1( "emUnits = %d\n", emUnits);
	dprintf2( "xx = %d, xy = %d\n", xx, xy);
	dprintf2( "yx = %d, yy = %d\n", yx, yy);
	dprintf1( "exp = %d\n", xxexp);

	dprintf3( "mppem = %d, mppemx = %d, mppemy = %d\n", mppem, mppemx, mppemy);

	if( grid_fitting > 0)
		calcCVT();

	status = TRAFO_APPLIED;
}


// get font extent
//### is there a way to avoid to go through all glyphs and interpret them?
//### would it have advantages to delay the scan line conversion?

void Rasterizer::getFontExtent( FontExtent* fe)
{
	if( status == NOT_READY)
		applyTransformation();
	if( status == FONT_DONE)
		return;

	int i;
	i = ttFont->os2Table->winAscent;
	fe->yWinAscent = ((yy * i) << xxexp) >> 11;
	i = ttFont->os2Table->winDescent;
	fe->yWinDescent = ((yy * i) << xxexp) >> 11;

	fe->xLeftMin	=
	fe->xRightMin	=
	fe->xAdvanceMin	=
	fe->yAdvanceMin	=
	fe->yAscentMin	=
	fe->yDescentMin	=
	fe->xBlackboxMin=
	fe->yBlackboxMin= INT_MAX;

	fe->xLeftMax	=
	fe->xRightMax	=
	fe->xAdvanceMax	=
	fe->yAdvanceMax	=
	fe->yAscentMax	=
	fe->yDescentMax	=
	fe->xBlackboxMax=
	fe->yBlackboxMax= INT_MIN;

//###	if( format & XFIXED) {
//		fe->width = ttFont->getMaxWidth( mppemx);
//		dX = (((fe->width-1) | (SCANLINEPAD-1)) + 1) >> 3;
//	}

	U8* buffer	= fe->buffer;
	U32 buflen	= fe->buflen;
	fe->bmplen	= 0;
	endBMP		= buffer + buflen;

	int numGlyphs	= ttFont->maxpTable->numGlyphs;
	buflen		-= numGlyphs * sizeof(CharInfo) + 4;
	*(U32*)buffer	= numGlyphs;

	CharInfo* ci	= (CharInfo*)(buffer + sizeof(void*));
	U8* firstBitmap	= (U8*)&ci[ numGlyphs];
	buffer		= firstBitmap;
	for( int glyphNo = 0; glyphNo < numGlyphs && buflen > 0;
		++glyphNo, ++ci) {
		GlyphMetrics* gm = &ci->gm;

		dprintf2( "buflen[%d] = %d\n", glyphNo, buflen);

		int length = putGlyphBitmap( glyphNo, buflen, buffer, gm);

		ci->length	= length;
		ci->offset	= buffer - firstBitmap;
		buffer		+= length;
		buflen		-= length;
		fe->bmplen	+= length;

#define MINMAX(x,y,z)	if( z < fe->x) fe->x = z;	\
			if( z > fe->y) fe->y = z;

		MINMAX( xBlackboxMin,	xBlackboxMax,	gm->xBlackbox);
		MINMAX( yBlackboxMin,	yBlackboxMax,	gm->yBlackbox);

		MINMAX( xLeftMin,	xLeftMax,	-gm->xOrigin);
		MINMAX( xRightMin,	xRightMax,	gm->xBlackbox-gm->xOrigin);

		MINMAX( yAscentMin,	yAscentMax,	gm->yBlackbox-gm->yOrigin);
		MINMAX( yDescentMin,	yDescentMax,	gm->yOrigin);

		MINMAX( xAdvanceMin,	xAdvanceMax,	gm->xAdvance);
		MINMAX( yAdvanceMin,	yAdvanceMax,	gm->yAdvance);
	}

	status = FONT_DONE;
}

int Rasterizer::putChar8Bitmap( char c, int size, U8* bmp, GlyphMetrics* gm)
{
	dprintf1( "charNo8 = %d", c);
	int glyphNo = ttFont->getGlyphNo8( c);
	return putGlyphBitmap( glyphNo, size, bmp, gm);
}


int Rasterizer::putChar16Bitmap( int c, int size, U8* bmp, GlyphMetrics* gm)
{
	dprintf1( "charNo16 = %d", c);
	int glyphNo = ttFont->getGlyphNo16( c);
	return putGlyphBitmap( glyphNo, size, bmp, gm);
}


int Rasterizer::putGlyphBitmap( int glyphNo, int size, U8* bmp, GlyphMetrics* gm)
{
	dprintf1( "\n=============== glyphNo %d ==================\n", glyphNo);

	if( size <= 0)
		return 0;

	GlyphTable* g = ttFont->glyphTable;
	if( !ttFont->getGlyphData( glyphNo, g)) {
		gm->xAdvance	= ttFont->getGlyphWidth( mppemx, glyphNo);
		gm->xOrigin	=
		gm->yOrigin	=
		gm->xBlackbox	=
		gm->yBlackbox	=
		gm->yAdvance	= 0;
		return (length = 0);
	}

	scaleGlyph( g, glyphNo);
	printOutline();

	int xmin = INT_MAX;
	int ymin = INT_MAX;
	int xmax = INT_MIN;
	int ymax = INT_MIN;

	register point* pp = p[ 1];
	for( int i = nPointsZ1; --i >= 0; ++pp) {
		if( MAGNIFY > 1) {
			pp->xnow *= MAGNIFY;
			pp->ynow *= MAGNIFY;
		}
		if( anti_aliasing > 0) {
			pp->xnow <<= anti_aliasing;
			pp->ynow <<= anti_aliasing;
		}
		int xt = pp->xnow;
		if( xmin > xt) xmin = xt;
		if( xmax < xt) xmax = xt;
		int yt = pp->ynow;
		if( ymin > yt) ymin = yt;
		if( ymax < yt) ymax = yt;
	}

	if( anti_aliasing > 0) {
		xmin &= -64 << anti_aliasing;
		ymin &= -64 << anti_aliasing;
		ymax |= 64;
	} else {
		xmin &= -64;
		ymin &= -64;
		xmax += 63;
		ymax += 63;
	}

	pp = p[1];
	nPointsZ1 += 2;
	for( int iNormalize = nPointsZ1; --iNormalize >= 0; ++pp) {
		pp->xnow -= xmin;
		pp->ynow -= ymin;
	}

	xmax >>= SHIFT;	ymax >>= SHIFT;
	xmin >>= SHIFT;	ymin >>= SHIFT;

	gm->xBlackbox	= width	= xmax - xmin + 1;
	gm->yBlackbox	= height= ymax - ymin + 1;

	--pp;
	gm->xAdvance	= pp->xnow >> SHIFT;
	gm->yAdvance	= pp->ynow >> SHIFT;
	--pp;
	gm->xOrigin	= pp->xnow >> SHIFT;
	gm->yOrigin	= pp->ynow >> SHIFT;

//###	if( (format & XFIXED) == 0)
		dX = (((width-1) | (SCANLINEPAD-1)) + 1) >> 3;
	length = dX * height;

	dprintf3( "width = %d, dX = %d, height = %d\n", width, dX, height);
	dprintf2( "gn=%d, length= %d\n", glyphNo, length);

	if( length >= size)
		return 0;

	currBMP = bmp;
	endBMP = bmp + size;
	memset( currBMP, 0, length);

	pp = p[ 1];
	for( int iOnlyCurve = nPointsZ1; --iOnlyCurve >= 0; ++pp)
		pp->flags &= ON_CURVE;

	for( int ci = 0, startPoint = 0; ci < nEndPoints; ++ci) {
		int endPoint = endPoints[ ci];
		drawContour( startPoint, endPoint);
		startPoint = endPoint + 1;
	}

	delete[] p[1];
	delete[] endPoints;

	if( anti_aliasing == 1)
		antiAliasing2();
	return length;
}


void Rasterizer::scaleGlyph( GlyphTable* g, int glyphNo)
{
	nEndPoints = g->nEndPoints;
	nPointsZ1 = g->nPoints;

	endPoints = g->endPoints;
	p[1] = g->points;

	register point* pp;
	int val;

	// scale outline
	pp = p[1];
	for( int i = nPointsZ1; --i >= 0; ++pp) {
		int xval = pp->xnow;
		int yval = pp->ynow;
		val = xx * xval + xy * yval;
		pp->xold = pp->xnow = ((val + 16) >> 5) << xxexp;
		val = yx * xval + yy * yval;
		pp->yold = pp->ynow = ((val + 16) >> 5) << xxexp;
#ifdef DEBUG
		pp->xgoal = pp->ygoal = 0;
#endif /* DEBUG */
	}

	// prepare phantom point 0
	int advanceWidth, leftSideBear;
	ttFont->hmtxTable->getHMetrics( glyphNo, &advanceWidth, &leftSideBear);
	val = g->xmin - leftSideBear;
	pp->xold = pp->xnow = ((xx * val + 16) >> 5) << xxexp;
	pp->yold = pp->ynow = ((yx * val + 16) >> 5) << xxexp;
	dprintf3( "xmin = %d, adv = %d, lsb = %d\n", g->xmin, advanceWidth, leftSideBear);
	pp->xnow &= -64;
	pp->ynow &= -64;
	//### ref implementation always has (0 0)!
	val = pp->xnow = pp->ynow = 0;	//###
	dprintf2( "phantom[0] = %5d -> %5d\n", val, pp->xold);

	// prepare phantom point 1
	val += advanceWidth;
	++pp;
	pp->xold = ((xx * val + 16) >> 5) << xxexp;
	pp->yold = ((yx * val + 16) >> 5) << xxexp;
	pp->xnow = (pp->xold + 32) & -64;
	pp->ynow = (pp->yold + 32) & -64;
	dprintf2( "phantom[1] = %5d -> %5d\n", val, pp->xold);


	switch( status) {
	case INVALID_FONT:
		return;
	case FONT_DONE:
		dprintf0( "### We shouldn't waste time like this ###\n");
		return;
	case NOT_READY:
		applyTransformation();
		// fall through
	case TRAFO_APPLIED:
		break;
	}

	if( grid_fitting <= 0 || g->codeLength == 0)
		return;

	dprintf1( "\n=== glyphCode ===, nPointsZ1 = %d", nPointsZ1);
//###	gs = default_gs;
gs.init( p);
gs.cvt_cut_in = default_gs.cvt_cut_in;
	gs.zp2 = gs.zp1 = gs.zp0 = p[1];
	//### why does g->codelength cause problems sometimes???
	stack = stackbase;
	execHints( g, g->codeOffset, g->codeLength);
}

void Rasterizer::printOutline( void)
{
	dprintf0( "\n=== grid fitted outline ===\n");
	point* pp = p[ 1];
	for( int i = 0, j = 0; i < nPointsZ1 + 2; ++i, ++pp) {
		dprintf3( "p[%d]\t%6d %6d  ", i, pp->xold, pp->yold);
		dprintf2( "-> %6d %6d", pp->xnow, pp->ynow);
		dprintf2( "  %d%d", (pp->flags & X_TOUCHED) != 0,
			(pp->flags & Y_TOUCHED) != 0);

		dprintf1( " %c", (pp->flags & ON_CURVE) ? '*' : ' ');

		dprintf2( "  (%6d %6d)", pp->xgoal, pp->ygoal);

		dprintf2( "  %c%c",
			(pp->xnow == pp->xgoal) ? '+' : '@',
			(pp->ynow == pp->ygoal) ? '+' : '@');

		dprintf2( "  %d %d\n",
			pp->xnow - pp->xgoal,
			pp->ynow - pp->ygoal);

		if( i == endPoints[ j]) {
			++j;
			dprintf0( "----\n");
		}
	 }
#if 0
	pp = &p[ 1][ nPointsZ1];
	fprintf( outfile, "phantom[0] = { %d, %d } <=> { %d, %d }\n",
		pp->xnow, pp->ynow, pp->xgoal, pp->ygoal);
	++pp;
	fprintf( outfile, "phantom[1] = { %d, %d}\t<=> { %d, %d}\t%d%d\n",
		pp->xnow, pp->ynow, pp->xgoal, pp->ygoal,
		pp->xnow==pp->xgoal, pp->ynow==pp->ygoal);
#endif
}

